/*!
  \file   io/db_cache/transaction_shrinker/DBCacheTransactionShrinker.hpp
  \brief  This is the interface class of a database cacher,
  which enables transcation shrinking.

  \author Ferenc BODON
  \date   2005-05-04
*/
#ifndef DBCacheTransactionShrinker_HPP
#define DBCacheTransactionShrinker_HPP

#include "common.hpp"


/**
   This is the interface class of a database cacher,
   which enables transcation shrinking.

   <p> In the basic database cacher (see DBCache) it is not possible
   to modify the transactions that are stored in the
   cacher. Nevertheless it may significantly increase the efficiency
   if we could delete some item from certain transactions. For example
   unnecessary items slow down the support count of Apriori regardless
   of the data structure that is used to store the candidates.  </p>

   <p> An efficient transaction skrinking, which can be used in
   Apriori, was proposed by Park et al. It states that an item can be
   deleted from transaction \a t at iteration \a l, if it is not the
   element of \a l different candidates that are contained in \a t.
   </p>   
 */
class DBCacheTransactionShrinker : public DBCache
{
   public:
      typedef typename DBCache::params_t params_t;
      DBCacheTransactionShrinker( const params_t* par ):
	 DBCache(par){}

/** It returns the next transaction and modifies the 
     previously returned transaction according to the filter.

     \param transaction the next transaction 

     \param filter the element of index \a i of the filter stores the
     number of candidates that contain item \a i and are contained in
     the previous transaction. According to the transaction shrinking
     proposed by Park et al. elements smaller than the size of the
     candidates can be removed from the transaction.
*/
      counter_t nextTransactionBIS_modifyPrevious( 
	 BIS& transaction, std::vector<unsigned int>& filter);

/// This sets the size of the candidates.
      void setCandidateSize(const size_t candidate_size);

};

#endif
